package top.ninwoo.utils.service.impl;

import com.spotify.docker.client.DockerClient;
import com.spotify.docker.client.exceptions.DockerException;
import com.spotify.docker.client.messages.Container;
import com.spotify.docker.client.messages.ContainerInfo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import top.ninwoo.utils.entity.DockerContainer;
import top.ninwoo.utils.service.DockerService;
import top.ninwoo.utils.service.LinuxCtlService;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

@Service
public class DockerServiceImpl implements DockerService {

    @Autowired
    private LinuxCtlService linuxService;

    @Autowired
    private DockerClient dockerClient;

    // 这里考虑要不要区分正在运行中的容器
    private ConcurrentHashMap<String, DockerContainer> containersMap = new ConcurrentHashMap<>();

    @Override
    public List<String> getDockerIds() {
        return getDockerIds(false);
    }
    /**
     * 获取容器Id
     * @param isAll 是否获取全部的容器id,包含未启动的容器
     * @return
     */
    @Override
    public List<String> getDockerIds(boolean isAll) {
        return new ArrayList<String>(containersMap.keySet());
    }

    /**
     * 将DockerClient获取的到结果封装成为我们自定义的数据结构
     * @param containers
     * @return
     */
    public List<DockerContainer> convertDockerResult(List<Container> containers) {
        if(containers == null) {
            throw new RuntimeException("容器列表不能为null");
        }
        List<DockerContainer> dockerContainers = new ArrayList<DockerContainer>(containers.size());
        containers.forEach(c -> {
            DockerContainer container = new DockerContainer();
            container.setId(c.id());
            container.setCommand(c.command());
            container.setCreated(c.created());
            container.setPorts(c.portsAsString());
            container.setImage(c.image());
            container.setStatus(c.status());
            container.setName(c.names().get(0));
            dockerContainers.add(container);
        });
        return dockerContainers;
    }

    @Override
    public void updateContainers() {
        updateContainers(true);
    }

    /**
     * 更新容器列表
     * @param isfull true:全量更新
     */
    @Override
    public void updateContainers(boolean isfull) {
        if(isfull) {
            // 清空
            containersMap.clear();
            List<Container> containers =
                    null;
            try {
                containers = dockerClient.listContainers(DockerClient.ListContainersParam.withStatusRunning());
            } catch (DockerException e) {
                e.printStackTrace();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            List<DockerContainer> dockerContainers = convertDockerResult(containers);
            dockerContainers.forEach(dockerContainer -> containersMap.put(dockerContainer.getId(), dockerContainer));
        } else {
            // TODO: 增量备份
        }
    }

    /**
     * 使用系统命令 docker run -itd 在后台启动一个容器,
     * 使用这种方式的一个重要愿意,是因为我没搞明白docker java启动容器的api到底如何使用
     * @param container 输入一个容器的描述信息
     * @return 返回容器启动之后真实的描述信息
     */
    @Override
    public DockerContainer runDocker(DockerContainer container) {
        // TODO: 这里启动Docker容器,需要再研究port如何起作用
        // TODO: 这里还需要处理name等属性为空的情况
        String cmd = "docker run -itd --name " + container.getName() + " " + container.getImage() + " " + container.getCommand();

        String result = linuxService.runCmd(cmd);
        if(result.contains("Error")) {
            throw new RuntimeException("Run Docker failed!:" + cmd);
        }
        // TODO:需要从Docker中查询处完整的信息
        container.setId(result);
        return container;
    }

    @Override
    public DockerContainer getDockerById(String id) {
        ContainerInfo containerInfo = null;
        try {
             containerInfo = dockerClient.inspectContainer(id);
        } catch (DockerException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return exchangeFromDockerInfo(containerInfo);
    }

    public DockerContainer exchangeFromDockerInfo(ContainerInfo containerInfo) {
        if(containerInfo == null) {
            throw new RuntimeException("ContainerInfo cannot be null");
        }
        DockerContainer dockerContainer = new DockerContainer();
        dockerContainer.setImage(containerInfo.image());
        dockerContainer.setName(containerInfo.name());
        dockerContainer.setId(containerInfo.id());
        dockerContainer.setStatus(containerInfo.state().status());
        dockerContainer.setCreated(containerInfo.created().getTime());
        return dockerContainer;
    }

    @Override
    public boolean deleteDockerById(String id) {
        try {
            dockerClient.stopContainer(id, 0);
            dockerClient.removeContainer(id);
            return true;
        } catch (DockerException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return false;
    }
}
